@@ -34,6 +34,10 @@ typedef struct {
 } SinkFunc;
 static GArray *sink_funcs = NULL;
 
+GHashTable *custom_type_registration = NULL;
+
+PyTypeObject *PyGObject_MetaType = NULL;
+
 /**
  * pygobject_sink:
  * @obj: a GObject
@@ -334,7 +338,7 @@ pygobject_props_get_iter(PyGProps *self)
     return (PyObject *) iter;
 }
 
-static int
+static Py_ssize_t
 PyGProps_length(PyGProps *self)
 {
     GObjectClass *class;
@@ -344,17 +348,17 @@ PyGProps_length(PyGProps *self)
     g_object_class_list_properties(class, &n_props);
     g_type_class_unref(class);
     
-    return (int)n_props;
+    return (Py_ssize_t)n_props;
 }
 
 static PySequenceMethods _PyGProps_as_sequence = {
-    (inquiry)PyGProps_length,
-    (binaryfunc)0,
-    (intargfunc)0,
-    (intargfunc)0,
-    (intintargfunc)0,
-    (intobjargproc)0,
-    (intintobjargproc)0
+    (lenfunc) PyGProps_length,
+    0,
+    0,
+    0,
+    0,
+    0,
+    0
 };
 
 PyTypeObject PyGProps_Type = {
@@ -478,9 +482,9 @@ pygobject_register_class(PyObject *dict, const gchar *type_name,
     PyObject *o;
     const char *class_name, *s;
     PyObject *runtime_bases;
-    PyObject *bases_list, *bases;
+    PyObject *bases_list, *bases, *mod_name;
     int i;
-
+    
     class_name = type->tp_name;
     s = strrchr(class_name, '.');
     if (s != NULL)
@@ -514,7 +518,7 @@ pygobject_register_class(PyObject *dict, const gchar *type_name,
     } else
         bases = runtime_bases;
 
-    type->ob_type = &PyGObject_MetaType;
+    type->ob_type = PyGObject_MetaType;
     type->tp_bases = bases;
     if (G_LIKELY(bases)) {
         type->tp_base = (PyTypeObject *)PyTuple_GetItem(bases, 0);
@@ -526,6 +530,16 @@ pygobject_register_class(PyObject *dict, const gchar *type_name,
 	return;
     }
 
+    /* Set type.__module__ to the name of the module,
+     * otherwise it'll default to 'gobject', see #376099
+     */
+    s = strrchr(type->tp_name, '.');
+    if (s != NULL) {
+	mod_name = PyString_FromStringAndSize(type->tp_name, (int)(s - type->tp_name));
+	PyDict_SetItemString(type->tp_dict, "__module__", mod_name);
+	Py_DECREF(mod_name);
+    }
+    
     if (gtype) {
 	o = pyg_type_wrapper_new(gtype);
 	PyDict_SetItemString(type->tp_dict, "__gtype__", o);
@@ -603,8 +617,7 @@ pyg_type_get_bases(GType gtype)
  * @gtype: the GType of the GObject subclass.
  *
  * Creates a new PyTypeObject from the given GType with interfaces attached in
- * bases. It will currently not filter out interfaces already implemented by
- * it parents.
+ * bases.
  *
  * Returns: a PyTypeObject for the new type or NULL if it couldn't be created
  */
@@ -635,7 +648,7 @@ pygobject_new_with_interfaces(GType gtype)
     PyDict_SetItemString(dict, "__doc__", pyg_object_descr_doc_get());
 
     /* generate the pygtk module name and extract the base type name */
-    gtype_name = (gchar *)g_type_name(gtype);
+    gtype_name = (gchar*)g_type_name(gtype);
     if (g_str_has_prefix(gtype_name, "Gtk")) {
 	mod_name = "gtk";
 	gtype_name += 3;
@@ -724,6 +737,10 @@ pygobject_lookup_class(GType gtype)
 
     if (gtype == G_TYPE_INTERFACE)
 	return &PyGInterface_Type;
+    
+    py_type = pyg_type_get_custom(g_type_name(gtype));
+    if (py_type)
+	return py_type;
 
     py_type = g_type_get_qdata(gtype, pygobject_class_key);
     if (py_type == NULL) {
@@ -738,9 +755,10 @@ pygobject_lookup_class(GType gtype)
 }
 
 /**
- * pygobject_new:
+ * pygobject_new_full:
  * @obj: a GObject instance.
  * @sink: whether to sink any floating reference found on the GObject.
+ * @g_class: the GObjectClass
  *
  * This function gets a reference to a wrapper for the given GObject
  * instance.  If a wrapper has already been created, a new reference
@@ -750,7 +768,7 @@ pygobject_lookup_class(GType gtype)
  * Returns: a reference to the wrapper for the GObject.
  */
 PyObject *
-pygobject_new_full(GObject *obj, gboolean sink)
+pygobject_new_full(GObject *obj, gboolean sink, gpointer g_class)
 {
     PyGObject *self;
 
@@ -765,7 +783,11 @@ pygobject_new_full(GObject *obj, gboolean sink)
 	Py_INCREF(self);
     } else {
 	/* create wrapper */
-	PyTypeObject *tp = pygobject_lookup_class(G_OBJECT_TYPE(obj));
+	PyTypeObject *tp;
+        if (g_class)
+            tp = pygobject_lookup_class(G_OBJECT_CLASS_TYPE(g_class));
+        else
+            tp = pygobject_lookup_class(G_OBJECT_TYPE(obj));
         /* need to bump type refcount if created with
            pygobject_new_with_interfaces(). fixes bug #141042 */
         if (tp->tp_flags & Py_TPFLAGS_HEAPTYPE)
@@ -791,10 +813,11 @@ pygobject_new_full(GObject *obj, gboolean sink)
     return (PyObject *)self;
 }
 
+
 PyObject *
 pygobject_new(GObject *obj)
 {
-    return pygobject_new_full(obj, TRUE);
+    return pygobject_new_full(obj, TRUE, NULL);
 }
 
 static void
@@ -997,7 +1020,7 @@ pygobject_init(PyGObject *self, PyObject *args, PyObject *kwargs)
     }
 
     if (kwargs) {
-	int pos = 0;
+	Py_ssize_t pos = 0;
 	PyObject *key;
 	PyObject *value;
 
@@ -1009,7 +1032,7 @@ pygobject_init(PyGObject *self, PyObject *args, PyObject *kwargs)
 	    pspec = g_object_class_find_property (class, key_str);
 	    if (!pspec) {
 		PyErr_Format(PyExc_TypeError,
-			     "gobject `%s' doesn't support property `%s'",
+			     "object of type `%s' doesn't support property `%s'",
 			     g_type_name(object_type), key_str);
 		goto cleanup;
 	    }
@@ -1072,8 +1095,9 @@ pygobject_get_property(PyGObject *self, PyObject *args)
     pspec = g_object_class_find_property(G_OBJECT_GET_CLASS(self->obj),
 					 param_name);
     if (!pspec) {
-	PyErr_SetString(PyExc_TypeError,
-			"the object does not support the given parameter");
+	PyErr_Format(PyExc_TypeError,
+		     "object of type `%s' does not have property `%s'",
+		     g_type_name(G_OBJECT_TYPE(self->obj)), param_name);
 	return NULL;
     }
     if (!(pspec->flags & G_PARAM_READABLE)) {
@@ -1104,8 +1128,9 @@ pygobject_set_property(PyGObject *self, PyObject *args)
     pspec = g_object_class_find_property(G_OBJECT_GET_CLASS(self->obj),
 					 param_name);
     if (!pspec) {
-	PyErr_SetString(PyExc_TypeError,
-			"the object does not support the given parameter");
+	PyErr_Format(PyExc_TypeError,
+		     "object of type `%s' does not have property `%s'",
+		     g_type_name(G_OBJECT_TYPE(self->obj)), param_name);
 	return NULL;
     }
     
@@ -1200,7 +1225,8 @@ pygobject_connect(PyGObject *self, PyObject *args)
 {
     PyObject *first, *callback, *extra_args;
     gchar *name;
-    guint handlerid, sigid, len;
+    guint sigid, len;
+    gulong handlerid;
     GQuark detail = 0;
     GClosure *closure;
 
@@ -1238,7 +1264,7 @@ pygobject_connect(PyGObject *self, PyObject *args)
     handlerid = g_signal_connect_closure_by_id(self->obj, sigid, detail,
 					       closure, FALSE);
     Py_DECREF(extra_args);
-    return PyInt_FromLong(handlerid);
+    return PyLong_FromUnsignedLong(handlerid);
 }
 
 static PyObject *
@@ -1246,7 +1272,9 @@ pygobject_connect_after(PyGObject *self, PyObject *args)
 {
     PyObject *first, *callback, *extra_args;
     gchar *name;
-    guint handlerid, sigid, len;
+    guint sigid;
+    gulong handlerid;
+    Py_ssize_t len;
     GQuark detail;
     GClosure *closure;
 
@@ -1285,7 +1313,7 @@ pygobject_connect_after(PyGObject *self, PyObject *args)
     handlerid = g_signal_connect_closure_by_id(self->obj, sigid, detail,
 					       closure, TRUE);
     Py_DECREF(extra_args);
-    return PyInt_FromLong(handlerid);
+    return PyLong_FromUnsignedLong(handlerid);
 }
 
 static PyObject *
@@ -1293,7 +1321,9 @@ pygobject_connect_object(PyGObject *self, PyObject *args)
 {
     PyObject *first, *callback, *extra_args, *object;
     gchar *name;
-    guint handlerid, sigid, len;
+    guint sigid;
+    gulong handlerid;
+    Py_ssize_t len;
     GQuark detail;
     GClosure *closure;
 
@@ -1332,7 +1362,7 @@ pygobject_connect_object(PyGObject *self, PyObject *args)
     handlerid = g_signal_connect_closure_by_id(self->obj, sigid, detail,
 					       closure, FALSE);
     Py_DECREF(extra_args);
-    return PyInt_FromLong(handlerid);
+    return PyLong_FromUnsignedLong(handlerid);
 }
 
 static PyObject *
@@ -1340,7 +1370,9 @@ pygobject_connect_object_after(PyGObject *self, PyObject *args)
 {
     PyObject *first, *callback, *extra_args, *object;
     gchar *name;
-    guint handlerid, sigid, len;
+    guint sigid;
+    gulong handlerid;
+    Py_ssize_t len;
     GQuark detail;
     GClosure *closure;
 
@@ -1379,15 +1411,15 @@ pygobject_connect_object_after(PyGObject *self, PyObject *args)
     handlerid = g_signal_connect_closure_by_id(self->obj, sigid, detail,
 					       closure, TRUE);
     Py_DECREF(extra_args);
-    return PyInt_FromLong(handlerid);
+    return PyLong_FromUnsignedLong(handlerid);
 }
 
 static PyObject *
 pygobject_disconnect(PyGObject *self, PyObject *args)
 {
-    guint handler_id;
+    gulong handler_id;
 
-    if (!PyArg_ParseTuple(args, "i:GObject.disconnect", &handler_id))
+    if (!PyArg_ParseTuple(args, "k:GObject.disconnect", &handler_id))
 	return NULL;
     
     CHECK_GOBJECT(self);
@@ -1400,9 +1432,9 @@ pygobject_disconnect(PyGObject *self, PyObject *args)
 static PyObject *
 pygobject_handler_is_connected(PyGObject *self, PyObject *args)
 {
-    guint handler_id;
+    gulong handler_id;
 
-    if (!PyArg_ParseTuple(args, "i:GObject.handler_is_connected", &handler_id))
+    if (!PyArg_ParseTuple(args, "k:GObject.handler_is_connected", &handler_id))
 	return NULL;
 
     
@@ -1414,9 +1446,9 @@ pygobject_handler_is_connected(PyGObject *self, PyObject *args)
 static PyObject *
 pygobject_handler_block(PyGObject *self, PyObject *args)
 {
-    guint handler_id;
+    gulong handler_id;
 
-    if (!PyArg_ParseTuple(args, "i:GObject.handler_block", &handler_id))
+    if (!PyArg_ParseTuple(args, "k:GObject.handler_block", &handler_id))
 	return NULL;
     
     CHECK_GOBJECT(self);
@@ -1429,9 +1461,9 @@ pygobject_handler_block(PyGObject *self, PyObject *args)
 static PyObject *
 pygobject_handler_unblock(PyGObject *self, PyObject *args)
 {
-    guint handler_id;
+    gulong handler_id;
 
-    if (!PyArg_ParseTuple(args, "i:GObject.handler_unblock", &handler_id))
+    if (!PyArg_ParseTuple(args, "k:GObject.handler_unblock", &handler_id))
 	return NULL;
     g_signal_handler_unblock(self->obj, handler_id);
     Py_INCREF(Py_None);
@@ -1441,7 +1473,8 @@ pygobject_handler_unblock(PyGObject *self, PyObject *args)
 static PyObject *
 pygobject_emit(PyGObject *self, PyObject *args)
 {
-    guint signal_id, i, len;
+    guint signal_id, i;
+    Py_ssize_t len;
     GQuark detail;
     PyObject *first, *py_ret;
     gchar *name;
@@ -1474,7 +1507,7 @@ pygobject_emit(PyGObject *self, PyObject *args)
 	gchar buf[128];
 
 	g_snprintf(buf, sizeof(buf),
-		   "%d parameters needed for signal %s; %d given",
+		   "%d parameters needed for signal %s; %ld given",
 		   query.n_params, name, len - 1);
 	PyErr_SetString(PyExc_TypeError, buf);
 	return NULL;
@@ -1504,7 +1537,8 @@ pygobject_emit(PyGObject *self, PyObject *args)
 	    g_free(params);
 	    return NULL;
 	}
-    }
+    }    
+
     if (query.return_type != G_TYPE_NONE)
 	g_value_init(&ret, query.return_type & ~G_SIGNAL_TYPE_STATIC_SCOPE);
     
@@ -1553,7 +1587,8 @@ static PyObject *
 pygobject_chain_from_overridden(PyGObject *self, PyObject *args)
 {
     GSignalInvocationHint *ihint;
-    guint signal_id, i, len;
+    guint signal_id, i;
+    Py_ssize_t len;
     PyObject *py_ret;
     const gchar *name;
     GSignalQuery query;
@@ -1581,7 +1616,7 @@ pygobject_chain_from_overridden(PyGObject *self, PyObject *args)
 	gchar buf[128];
 
 	g_snprintf(buf, sizeof(buf),
-		   "%d parameters needed for signal %s; %d given",
+		   "%d parameters needed for signal %s; %ld given",
 		   query.n_params, name, len);
 	PyErr_SetString(PyExc_TypeError, buf);
 	return NULL;
@@ -1629,6 +1664,102 @@ pygobject_chain_from_overridden(PyGObject *self, PyObject *args)
     return py_ret;
 }
 
+static PyObject *
+pygobject_disconnect_by_func(PyGObject *self, PyObject *args)
+{
+    PyObject *pyfunc = NULL;
+    GClosure *closure = NULL;
+    guint retval;
+    
+    CHECK_GOBJECT(self);
+
+    if (!PyArg_ParseTuple(args, "O:GObject.disconnect_by_func", &pyfunc))
+	return NULL;
+
+    if (!PyCallable_Check(pyfunc)) {
+	PyErr_SetString(PyExc_TypeError, "first argument must be callable");
+	return NULL;
+    }
+
+    closure = gclosure_from_pyfunc(self, pyfunc);
+    if (!closure) {
+	PyErr_Format(PyExc_TypeError, "nothing connected to %s",
+		     PyString_AsString(PyObject_Repr((PyObject*)pyfunc)));
+	return NULL;
+    }
+    
+    retval = g_signal_handlers_disconnect_matched(self->obj,
+						  G_SIGNAL_MATCH_CLOSURE,
+						  0, 0,
+						  closure,
+						  NULL, NULL);
+    return PyInt_FromLong(retval);
+}
+
+static PyObject *
+pygobject_handler_block_by_func(PyGObject *self, PyObject *args)
+{
+    PyObject *pyfunc = NULL;
+    GClosure *closure = NULL;
+    guint retval;
+    
+    CHECK_GOBJECT(self);
+
+    if (!PyArg_ParseTuple(args, "O:GObject.handler_block_by_func", &pyfunc))
+	return NULL;
+
+    if (!PyCallable_Check(pyfunc)) {
+	PyErr_SetString(PyExc_TypeError, "first argument must be callable");
+	return NULL;
+    }
+
+    closure = gclosure_from_pyfunc(self, pyfunc);
+    if (!closure) {
+	PyErr_Format(PyExc_TypeError, "nothing connected to %s",
+		     PyString_AsString(PyObject_Repr((PyObject*)pyfunc)));
+	return NULL;
+    }
+    
+    retval = g_signal_handlers_block_matched(self->obj,
+					     G_SIGNAL_MATCH_CLOSURE,
+					     0, 0,
+					     closure,
+					     NULL, NULL);
+    return PyInt_FromLong(retval);
+}
+
+static PyObject *
+pygobject_handler_unblock_by_func(PyGObject *self, PyObject *args)
+{
+    PyObject *pyfunc = NULL;
+    GClosure *closure = NULL;
+    guint retval;
+    
+    CHECK_GOBJECT(self);
+
+    if (!PyArg_ParseTuple(args, "O:GObject.handler_unblock_by_func", &pyfunc))
+	return NULL;
+
+    if (!PyCallable_Check(pyfunc)) {
+	PyErr_SetString(PyExc_TypeError, "first argument must be callable");
+	return NULL;
+    }
+
+    closure = gclosure_from_pyfunc(self, pyfunc);
+    if (!closure) {
+	PyErr_Format(PyExc_TypeError, "nothing connected to %s",
+		     PyString_AsString(PyObject_Repr((PyObject*)pyfunc)));
+	return NULL;
+    }
+    
+    retval = g_signal_handlers_unblock_matched(self->obj,
+					       G_SIGNAL_MATCH_CLOSURE,
+					       0, 0,
+					       closure,
+					       NULL, NULL);
+    return PyInt_FromLong(retval);
+}
+
 static PyMethodDef pygobject_methods[] = {
     { "__gobject_init__", (PyCFunction)pygobject__gobject_init__,
       METH_VARARGS|METH_KEYWORDS },
@@ -1644,10 +1775,13 @@ static PyMethodDef pygobject_methods[] = {
     { "connect_object", (PyCFunction)pygobject_connect_object, METH_VARARGS },
     { "connect_object_after", (PyCFunction)pygobject_connect_object_after, METH_VARARGS },
     { "disconnect", (PyCFunction)pygobject_disconnect, METH_VARARGS },
+    { "disconnect_by_func", (PyCFunction)pygobject_disconnect_by_func, METH_VARARGS },
     { "handler_disconnect", (PyCFunction)pygobject_disconnect, METH_VARARGS },
     { "handler_is_connected", (PyCFunction)pygobject_handler_is_connected, METH_VARARGS },
     { "handler_block", (PyCFunction)pygobject_handler_block, METH_VARARGS },
     { "handler_unblock", (PyCFunction)pygobject_handler_unblock,METH_VARARGS },
+    { "handler_block_by_func", (PyCFunction)pygobject_handler_block_by_func, METH_VARARGS },
+    { "handler_unblock_by_func", (PyCFunction)pygobject_handler_unblock_by_func, METH_VARARGS },
     { "emit", (PyCFunction)pygobject_emit, METH_VARARGS },
     { "stop_emission", (PyCFunction)pygobject_stop_emission, METH_VARARGS },
     { "emit_stop_by_name", (PyCFunction)pygobject_stop_emission,METH_VARARGS },
@@ -1727,91 +1861,3 @@ PyTypeObject PyGObject_Type = {
 };
 
 
-static int
-pygobjectmeta_register(PyTypeObject *subtype)
-{
-    PyObject *pytype_name;
-    const char *type_name = NULL;
-    int retval = 0;
-    
-    /* Maybe the type doesn't really need to registered? */
-    if (subtype->tp_dict == NULL)
-	goto out;
-	
-    if (!(PyDict_GetItemString(subtype->tp_dict, "__gtype_name__")
-          || PyDict_GetItemString(subtype->tp_dict, "__gproperties__")
-          || PyDict_GetItemString(subtype->tp_dict, "__gsignals__")))
-	goto out;
-
-    /* If it's already registered, skip registration */
-    if (PyDict_GetItemString(subtype->tp_dict, "__gtype__"))
-	goto out;
-
-    pytype_name = PyDict_GetItemString(subtype->tp_dict, "__gtype_name__");
-    if (pytype_name)
-	type_name = PyString_AsString(pytype_name);
-    retval = pyg_type_register(subtype, type_name);
-
-out:
-    return retval;
-}
-
-/* -------- GObject MetaClass -----------*/
-static int
-pygobjectmeta_init(PyTypeObject *subtype, PyObject *args, PyObject *kwargs)
-{
-    if (PyType_Type.tp_init((PyObject *) subtype, args, kwargs))
-        return -1;
-    
-    return pygobjectmeta_register(subtype);
-}
-
-
-PyTypeObject PyGObject_MetaType = {
-    PyObject_HEAD_INIT(NULL)
-    0,					/* ob_size */
-    "gobject.GObjectMeta",		/* tp_name */
-    0,					/* tp_basicsize */
-    0,					/* tp_itemsize */
-    /* methods */
-    (destructor)0,			/* tp_dealloc */
-    (printfunc)0,			/* tp_print */
-    (getattrfunc)0,			/* tp_getattr */
-    (setattrfunc)0,			/* tp_setattr */
-    (cmpfunc)0,				/* tp_compare */
-    (reprfunc)0,			/* tp_repr */
-    0,					/* tp_as_number */
-    0,					/* tp_as_sequence */
-    0,					/* tp_as_mapping */
-    (hashfunc)0,			/* tp_hash */
-    (ternaryfunc)0,			/* tp_call */
-    (reprfunc)0,			/* tp_str */
-    (getattrofunc)0,			/* tp_getattro */
-    (setattrofunc)0,			/* tp_setattro */
-    0,					/* tp_as_buffer */
-    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_BASETYPE, /* tp_flags */
-    "GObject Metaclass -- automatically registers GObject subclasses "
-    "as new GType's", /* Documentation string */
-    (traverseproc)0,			/* tp_traverse */
-    (inquiry)0,				/* tp_clear */
-    (richcmpfunc)0,			/* tp_richcompare */
-    0,					/* tp_weaklistoffset */
-    (getiterfunc)0,			/* tp_iter */
-    (iternextfunc)0,			/* tp_iternext */
-    0,                                  /* tp_methods */
-    0,					/* tp_members */
-    0,					/* tp_getset */
-    NULL,				/* tp_base */
-    (PyObject *)0,			/* tp_dict */
-    0,					/* tp_descr_get */
-    0,					/* tp_descr_set */
-    0,					/* tp_dictoffset */
-    (initproc)pygobjectmeta_init,	/* tp_init */
-    (allocfunc)0,			/* tp_alloc */
-    (newfunc)0,				/* tp_new */
-    0,					/* tp_free */
-    (inquiry)0,				/* tp_is_gc */
-    NULL,				/* tp_bases */
-};
-
-
