@@ -293,30 +293,32 @@ struct _PyGObject_Functions *_PyGObject_API;
     } \
 } G_STMT_END
 
-#define init_pygobject_check(major, minor, micro) G_STMT_START {          \
-    init_pygobject();                                                     \
-    PyObject *gobject = PyImport_ImportModule("gobject");                 \
-    PyObject *mdict = PyModule_GetDict(gobject);                          \
-    PyObject *version = PyDict_GetItemString(mdict, "pygobject_version"); \
-    int found_major, found_minor, found_micro;                            \
-    if (!version)                                                         \
-        version = PyDict_GetItemString(mdict, "pygtk_version");           \
-    if (!version) {                                                       \
-        PyErr_SetString(PyExc_ImportError,                                \
-                         "PyGObject version too old");                    \
-        return;                                                           \
-    }                                                                     \
-    if (!PyArg_ParseTuple(version, "iii",                                 \
-                          &found_major, &found_minor, &found_micro))      \
-        return;                                                           \
-    if (major > found_major || minor > found_minor ||                     \
-        micro > found_micro) {                                            \
-        PyErr_Format(PyExc_ImportError,                                   \
-                     "PyGObject version too old, %d.%d.%d is required, "  \
-                     "found %d.%d.%d.", major, minor, micro,              \
-                     found_major, found_minor, found_micro);              \
-        return;                                                           \
-    }                                                                     \
+#define init_pygobject_check(req_major, req_minor, req_micro) G_STMT_START { \
+    PyObject *gobject, *mdict, *version;                                     \
+    int found_major, found_minor, found_micro;                               \
+    init_pygobject();                                                        \
+    gobject = PyImport_ImportModule("gobject");                              \
+    mdict = PyModule_GetDict(gobject);                                       \
+    version = PyDict_GetItemString(mdict, "pygobject_version");              \
+    if (!version)                                                            \
+        version = PyDict_GetItemString(mdict, "pygtk_version");              \
+    if (!version) {                                                          \
+        PyErr_SetString(PyExc_ImportError,                                   \
+                         "PyGObject version too old");                       \
+        return;                                                              \
+    }                                                                        \
+    if (!PyArg_ParseTuple(version, "iii",                                    \
+                          &found_major, &found_minor, &found_micro))         \
+        return;                                                              \
+    if (req_major != found_major ||                                          \
+        req_minor >  found_minor ||                                          \
+        (req_minor == found_minor && req_micro > found_micro)) {	     \
+        PyErr_Format(PyExc_ImportError,                                      \
+                     "PyGObject version mismatch, %d.%d.%d is required, "    \
+                     "found %d.%d.%d.", req_major, req_minor, req_micro,     \
+                     found_major, found_minor, found_micro);                 \
+        return;                                                              \
+    }                                                                        \
 } G_STMT_END
 
 #endif /* !_INSIDE_PYGOBJECT_ */
