@@ -139,6 +139,16 @@ class E(gobject.GObject):
         assert self.status == 0
         self.status = 1
 
+class F(gobject.GObject):
+    __gsignals__ = { 'signal': (gobject.SIGNAL_RUN_FIRST, gobject.TYPE_NONE,
+                                ()) }
+    def __init__(self):
+        gobject.GObject.__init__(self)
+        self.status = 0
+
+    def do_signal(self):
+        self.status += 1
+
 class TestEmissionHook(unittest.TestCase):
     def testAdd(self):
         self.hook = True
@@ -168,6 +178,41 @@ class TestEmissionHook(unittest.TestCase):
             self.assertEqual(e.status, 1)
         e.status = 3
 
+    def testCallbackReturnFalse(self):
+        self.hook = False
+        obj = F()
+        def _emission_hook(obj):
+            obj.status += 1
+            return False
+        hook_id = gobject.add_emission_hook(obj, "signal", _emission_hook)
+        obj.emit('signal')
+        obj.emit('signal')
+        self.assertEqual(obj.status, 3)
+
+    def testCallbackReturnTrue(self):
+        self.hook = False
+        obj = F()
+        def _emission_hook(obj):
+            obj.status += 1
+            return True
+        hook_id = gobject.add_emission_hook(obj, "signal", _emission_hook)
+        obj.emit('signal')
+        obj.emit('signal')
+        gobject.remove_emission_hook(obj, "signal", hook_id)
+        self.assertEqual(obj.status, 4)
+
+    def testCallbackReturnTrueButRemove(self):
+        self.hook = False
+        obj = F()
+        def _emission_hook(obj):
+            obj.status += 1
+            return True
+        hook_id = gobject.add_emission_hook(obj, "signal", _emission_hook)
+        obj.emit('signal')
+        gobject.remove_emission_hook(obj, "signal", hook_id)
+        obj.emit('signal')
+        self.assertEqual(obj.status, 3)
+
 class TestClosures(unittest.TestCase):
     def setUp(self):
         self.count = 0
@@ -197,6 +242,23 @@ class TestClosures(unittest.TestCase):
         e.emit('signal')
         self.assertEqual(self.count, 1)
 
+    def testHandlerBlockMethod(self):
+        # Filed as #375589
+        class A:
+            def __init__(self):
+                self.a = 0
+
+            def callback(self, o):
+                self.a = 1
+                o.handler_block_by_func(self.callback)
+
+        inst = A()
+        e = E()
+        e.connect("signal", inst.callback)
+        e.emit('signal')
+        self.assertEqual(inst.a, 1)
+        gc.collect()
+
     def testGString(self):
         class C(gobject.GObject):
             __gsignals__ = { 'my_signal': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_GSTRING,
@@ -211,7 +273,35 @@ class TestClosures(unittest.TestCase):
         c = C(self)
         data = c.emit("my_signal", "\01\00\02")
         self.assertEqual(data, "\02\00\01")
-        
+
+class SigPropClass(gobject.GObject):
+    __gsignals__ = { 'my_signal': (gobject.SIGNAL_RUN_FIRST, gobject.TYPE_NONE,
+                                   (gobject.TYPE_INT,)) }
+
+    __gproperties__ = {
+        'foo': (str, None, None, '', gobject.PARAM_WRITABLE|gobject.PARAM_CONSTRUCT),
+        }
+
+    signal_emission_failed = False
+
+    def do_my_signal(self, arg):
+        self.arg = arg
+
+    def do_set_property(self, pspec, value):
+        if pspec.name == 'foo':
+            self._foo = value
+        else:
+            raise AttributeError, 'unknown property %s' % pspec.name
+        try:
+            self.emit("my-signal", 1)
+        except TypeError:
+            self.signal_emission_failed = True
+
+
+class TestSigProp(unittest.TestCase):
+    def testEmitInPropertySetter(self):
+        obj = SigPropClass()
+        self.failIf(obj.signal_emission_failed)
 
 if __name__ == '__main__':
     unittest.main()
